@@ -199,6 +199,10 @@ DEFAULT_FIELDS = {
         "TODO: adjust to your case.",
 }
 
+POPULATE_INTENDED_FOR_OPTS = {
+    'matching_parameters': ['ImagingVolume', 'Shims'],
+    'criterion': 'Closest'
+}
 
 def _delete_chars(from_str, deletechars):
     """ Delete characters from string allowing for Python 2 / 3 difference
@@ -547,7 +551,7 @@ def infotodict(seqinfo):
                                   if isinstance(current_run, int)
                                   else current_run)
         else:
-            # if there is no _run -- no run label addded
+            # if there is no _run -- no run label added
             run_label = None
 
         # yoh: had a wrong assumption
